local M = {}

local scenario = require("scenario/scenarios")
local scoreStats = require("statistics/statistics")
local imguiUtils = require('ui/imguiUtils')
local im = ui_imgui

local player, racers, racePath, stats, finishers = {}, {}, {}, {}, {}
local mapNodes, mapGp, aiLua, startPos, finishPos, finishRadius = nil, nil, nil, nil, nil, nil
local textBoxQueue = {}
local textBoxTimer = 0
local mainTimer = 0
local spawnDelay = 0
local hornValue = 0
local hornFlag = false
local up = vec3(0, 0, 1)
local carLimit = Engine.Platform.getCPUInfo().coresPhysical or 4
local userCarLimit = core_settings_settings.getValue("trafficAmount")

local ranks = {"1st", "2nd", "3rd", "4th", "5th", "6th", "7th", "8th", "9th", "10th"}
local event = {
  series = "default",
  index = 1,
  racesTotal = 1,
  circuit = false,
  aiLua = "ai",
  autoStart = nil,
  damageCutoff = 10000,
  stuckTime = 3,
  crawlSpeed = 3,
  startBoostDist = 100,
  aggroValue = 1.2,
  aggroVariance = 0,
  aggroTurnWatch = 0.9,
  aggroTurnFactor = 0,
  aggroTurnCutoff = 0.7,
  aggroReboundRate = 0.05,
  aggroRubberband = 0,
  avoidCars = false,
  textBoxTime = 7.5,
  playerName = "Player",
  racerNames = {},
  messages = {win = "You won the race!", finish = "won the race.", place = "You placed", stuck = "You got stuck!", flip = "You flipped over!"},
  rewards = {500, 250, 125, 60, 40, 20},
  saveFile = "settings/raceExtrasSave.json",
  achievementsFile = ""
}
local spawn = {
  group = nil,
  quantity = 1,
  order = true,
  mode = "raceGridAlt",
  gap = 15
}

local windowOpen = im.BoolPtr(false)
local baseColor = im.ImVec4(1, 1, 1, 1)
local activeColor = im.ImVec4(1, 0.55, 0.1, 1)
local achievementTex = imguiUtils.texObj("/art/sideAchievement.png")
local infoTex = imguiUtils.texObj("/art/sideInfo.png")
local headingTypes = {achievement = achievementTex, info = infoTex}
local styles = {}
styles[10] = im.ImVec4(0.8, 0.3216, 0, 1)
styles[11] = im.ImVec4(0.8, 0.3216, 0, 1)
styles[21] = im.ImVec4(1, 0.4039, 0, 1)
styles[22] = im.ImVec4(1, 0.5216, 0.2, 1)
styles[23] = im.ImVec4(0.8, 0.3216, 0, 1)

local function timeToString(t) -- returns a user friendly time string
  local minutes = math.floor(t / 60);
  local seconds = t - (minutes * 60);
  return string.format("%02.0f:%05.2f", minutes, seconds)
end

local function setupSaveFile() -- creates and returns blank save data for use as json
  local json = {
    money = 0,
    wins = 0,
    racesTotal = event.racesTotal or 1,
    raceScores = {},
    achievements = {},
    allowAchievements = true,
    series = event.series
  }
  for i = 1, json.racesTotal do
    table.insert(json.raceScores, 0)
  end
  json.achievements = {
    notDone = {},
    done = {}
  }
  local goals = jsonReadFile(event.achievementsFile)
  if goals then
    for i, v in ipairs(goals) do
      table.insert(json.achievements.notDone, v.name)
    end
  else
    json.allowAchievements = false
  end
  return json
end

local function getCondition(cond) -- returns true if an achievement condition statement checks out
  local valid = false
  local var1 = type(cond[1]) == "number" and cond[1] or stats[cond[1]]
  local var2 = type(cond[3]) == "number" and cond[3] or stats[cond[3]]
  if var1 and var2 then
    if cond[2] == "lte" then
      if var1 <= var2 then
        valid = true
      end
    elseif cond[2] == "gte" then
      if var1 >= var2 then
        valid = true
      end
    else
      if var1 == var2 then
        valid = true
      end
    end
  end
  return valid
end

local function testAchievements(json) -- checks and saves achievements
  local goals = jsonReadFile(event.achievementsFile)
  if goals then
    stats.achievementsDone = 0
    stats.achievementsTotal = #goals
    if stats.achievementsTotal == 0 then
      json.allowAchievements = false
    else
      for k, v in pairs(goals) do
        if getCondition(v.cond) then
          if not tableContains(json.achievements.done, v.name) then
            local num = nil
            for i, w in ipairs(json.achievements.notDone) do
              if w == v.name then
                num = i
              end
            end
            if num then
              table.remove(json.achievements.notDone, num)
              table.insert(json.achievements.done, v.name)
              local data = {}
              data.type = "achievement"
              data.title = v.title
              data.desc = v.desc
              table.insert(textBoxQueue, data)
            end
          end
        end
      end
      stats.achievementsDone = #json.achievements.done
    end
  end
end

local function showTextBox(data) -- shows a pop-up UI in the corner
  if not windowOpen[0] then return end
  im.SetNextWindowPos(im.ImVec2(0, 60), im.Cond_Always)
  im.SetNextWindowSize(im.ImVec2(math.max(400, string.len(data.desc) * 10), math.max(36, tableSize(data) * 36)), im.Cond_Always)
  for k, v in pairs(styles) do
    im.PushStyleColor2(k, v)
  end
  if im.Begin("Race Extras ", windowOpen, im.flags(im.WindowFlags_NoTitleBar, im.WindowFlags_NoResize)) then
    im.SetWindowFontScale(1.4)
    if data.type and headingTypes[data.type] then
      im.Image(headingTypes[data.type].texId, im.ImVec2(256, 24), im.ImVec2(0, 0), im.ImVec2(1, 1), im.GetStyleColorVec4(im.Col_Text))
    end
    if data.title then
      im.TextColored(activeColor, data.title)
    end
    if data.desc then
      im.Text(data.desc)
    end
    if data.extra then
      im.Text(data.extra)
    end
    im.End()
  end
  for k, v in pairs(styles) do
    im.PopStyleColor()
  end
end

local function setHornValue(value) -- sets the player horn value
  if value then hornValue = tonumber(value) end
end

local function createRacePath(pos) -- creates the entire race path, as well as the modified AI path
  if not mapNodes then mapNodes = map.getMap().nodes end
  if not mapGp then mapGp = map.getGraphpath() end
	local n1, n2 = nil, nil
  event.lapCountAi = event.circuit and event.lapCount + 1 or event.lapCount
  event.waypointsAi = deepcopy(event.waypoints)
  if not event.circuit then
    local num = #event.waypoints
    n1, n2 = event.waypoints[num - 1], event.waypoints[num]
    local shortPath = mapGp:getPath(n1, n2, 1)
    num = #shortPath
    local extendPath = mapGp:getRandomPathG(event.waypoints[#event.waypoints], (mapNodes[shortPath[num]].pos - mapNodes[shortPath[num - 1]].pos):z0():normalized(), 200)
    table.insert(event.waypointsAi, extendPath[#extendPath])
  end
  n1 = event.autoStart or map.findClosestRoad(pos)
	for i, v in ipairs(event.waypoints) do
		n2 = v
		local shortPath = mapGp:getPath(n1, n2, 1)
		for o, w in ipairs(shortPath) do
			if o > 1 then
				table.insert(racePath, w)
			end
		end
		n1 = n2
	end
end

local function placePlayer() -- automatically places the player at the defined start node
  local n1, n2 = event.autoStart, event.waypoints[1]
  local shortPath = map.getGraphpath():getPath(n1, n2, 1)
  local p1, p2 = map.getMap().nodes[shortPath[1]].pos, map.getMap().nodes[shortPath[2]].pos
  local pos = p1 + vec3(0, 0, 0.4)
  local rot = (p1 - p2):normalized()
  rot = quatFromDir(vec3(0, 1, 0):rotated(quatFromDir(rot, up)), up)
  be:getPlayerVehicle(0):setPositionRotation(pos.x, pos.y, pos.z, rot.x, rot.y, rot.z, rot.w)
end

local function setupPlayer() -- sets up player
  player.obj = be:getPlayerVehicle(0)
	player.name = player.obj.name
	player.id = player.obj:getID()
  player.pos = vec3(player.obj:getPosition())
  player.role = "player"
	player.path = {node = "", idx = 0, percent = 0}
	player.timers = {stuck = 0, flip = 0}
	player.lapsDone = 0
  player.state = "start"
  if event.playerName then
    core_vehicles.setPlateText(event.playerName, player.id)
    player.driver = event.playerName
  end
end

local function setupRacers() -- sets up racers
  local quantity = spawn.group and spawn.quantity or 20
	local others = extensions.core_multiSpawn.getNewVehicles(quantity)
	for i, v in ipairs(others) do
		local a = {}
		a.obj = scenetree.findObjectById(v)
		a.name = a.obj.name
		a.id = v
    a.pos = vec3(a.obj:getPosition())
    a.role = "ai"
    a.aggroBase = event.aggroValue + (math.random() * 2 - 1) * event.aggroVariance
    a.aggro = a.aggroBase
		a.path = {node = "", idx = 0, percent = 0}
		a.lapsDone = 0
    a.aggroPathIdx = 3
    a.aggroFactor = 1
    a.aggroAlt = a.aggro
    a.distTarget = vec3(0, 0, 0)
		a.state = "start"
    if event.racerNames[i] then
      core_vehicles.setPlateText(event.racerNames[i], a.id)
      a.driver = event.racerNames[i]
    end
    if event.scriptAi and event.scriptAi.files then
      for o, w in ipairs(event.scriptAi.files) do
        if (a.driver and string.find(a.driver, w)) or string.find(a.name, w) then
          a.script = w
        end
      end
    end
    if aiLua ~= "ai" then
      a.obj:queueLuaCommand("extensions.unload('"..aiLua.."')")
      a.obj:queueLuaCommand("extensions.load('"..aiLua.."')")
    end
		table.insert(racers, a)
	end
end

local function adjustAggro(v) -- dynamically adjusts the aggression value of an AI racer, to help improve its ability
  if event.aggroTurnWatch > 0 and event.aggroTurnFactor > 0 and racePath[v.path.idx + 12] then
    local newAggro = 1
    local newIdx = 1
    local newVel = math.pow(v.vel, 1.2)
    local bounds = {square(20 + newVel), square(40 + newVel)}
    local pSet = {}
    for i = 1, 8 do
      local nodeDist = v.pos:squaredDistance(mapNodes[racePath[v.path.idx + i]].pos)
      if nodeDist >= bounds[1] and nodeDist <= bounds[2] then
        for o = 1, 5 do
          pSet[o] = mapNodes[racePath[v.path.idx + i + o - 1]].pos
        end
        newIdx = i
        break
      end
    end
    if #pSet > 0 then
      local dotTotal = 0
      local height = 0
      local distTotal = pSet[1]:squaredDistance(pSet[2])
      local firstVec = (pSet[2] - pSet[1]):z0():normalized()
      local loop = 0
      for i = 3, 5 do
        dotTotal = dotTotal + firstVec:dot((pSet[i] - pSet[1]):z0():normalized())
        distTotal = distTotal + (pSet[i - 1]:squaredDistance(pSet[i]))
        height = math.max(0, pSet[1].z - pSet[i].z)
        loop = loop + 1
        if distTotal >= square(80) then break end
      end
      local dotAvg = math.max(0.01, dotTotal / loop)
      if dotAvg <= event.aggroTurnWatch then
        dotAvg = dotAvg - (height * (0.5 + newVel / 250)) / math.sqrt(distTotal)
        newAggro = v.aggro * clamp(dotAvg * event.aggroTurnFactor, event.aggroTurnCutoff, 1)
        if dotAvg < v.aggroFactor then
          v.aggroPathIdx = v.path.idx + newIdx + 1
          v.aggroFactor = dotAvg
          v.aggroAlt = newAggro
          v.obj:queueLuaCommand(aiLua.."setAggression("..tostring(newAggro)..")")
        end
      end
    end
  end
  if v.path.idx >= v.aggroPathIdx then
    v.aggroAlt = v.aggroAlt + event.aggroReboundRate
    local newAggro = math.min(v.aggro, v.aggroAlt)
    v.obj:queueLuaCommand(aiLua.."setAggression("..tostring(newAggro)..")")
    v.aggroFactor = 1
  end
end

local function pathProgress(v) -- tracks the race path progress of the given vehicle
  local bestDist = math.huge
  local newNode = nil
  for i = math.max(1, v.path.idx - 5), math.min(#racePath, v.path.idx + 5), 1 do
    local nodeDist = v.pos:squaredDistance(mapNodes[racePath[i]].pos)
    if nodeDist < bestDist then
      bestDist = nodeDist
      newNode = {node = racePath[i], idx = i, percent = (i / #racePath) * 100 + (100 * v.lapsDone)}
    end
  end
  if newNode then
    v.path = newNode
  end
  if v.role == "ai" and v.state == "race" then
    adjustAggro(v)
  end
end

local function insertNewPaths(data) -- inserts custom road segments into the map and reloads the map
  local missionFile = getMissionFilename()
  local dir, name, ext = path.split(missionFile, "(.-)([^/]-([^%.]*))$")
  local mapFile = dir.."map.json"
  local jsonMain = jsonReadFile(mapFile)
  if jsonMain then
    if not data.segments then
      data.segments = {}
    end
    if data.fileName then
      local json = jsonReadFile(data.fileName)
      if json then
        for k, v in pairs(json.segments) do
          data.segments[k] = v
        end
      end
    end
    for k, v in pairs(data.segments) do
      jsonMain.segments[k] = v
    end
    jsonWriteFile(dir.."map.json", jsonMain, true)
    map.reset()
    os.remove(dir.."map.json")
  end
end

local function spawnGroup() -- spawns a vehicle group
  if spawn.minQuantity then
    carLimit = carLimit - 1
    if userCarLimit and userCarLimit > 0 then
      carLimit = userCarLimit
    end
    spawn.quantity = clamp(carLimit, spawn.minQuantity, spawn.quantity)
  end
  extensions.core_multiSpawn.deleteVehicles(spawn.quantity)
  extensions.core_multiSpawn.setGroup("custom", spawn.group)
  extensions.core_multiSpawn.setupVehicles(spawn.quantity, spawn.order, spawn.mode, spawn.gap)
end

local function initScenario() -- initializes the scenario; gets called by onScenarioLoaded
  local this = scenario.getScenario()
  event.waypoints = this.lapConfig
  event.lapCount = this.lapCount or 1
  event.lapPoint = event.waypoints[#event.waypoints]
  if this.manualPaths then
    insertNewPaths(this.manualPaths)
  end
  if this.spawn then
    spawn = this.spawn
  end
  if this.raceExtras then
    for k, v in pairs(this.raceExtras) do
      if type(v) ~= "nil" then
        event[k] = v
      end
    end
  end
  if this.scriptAi then
    event.scriptAi = this.scriptAi
  end
  if event.autoStart then
    placePlayer()
  end
  if spawn.group then
    if spawnDelay then
      spawnDelay = 1
    else
      spawnGroup()
    end
	end
end

local function onScenarioUIReady()
  stats = {topSpeed = 0, topDamage = 0, maxRoll = 1, headStart = 0, racersCount = 0, finishTime = 0, finishPlace = 0, stuckRacers = {}, rolledRacers = {}}
	player, racers, finishers = {}, {}, {}
  mainTimer = 0
  aiLua = event.aiLua
  if event.autoStart then
    placePlayer()
  end
  startPos = vec3(be:getPlayerVehicle(0):getPosition())
  finishPos = vec3(scenetree.findObject(event.waypoints[#event.waypoints]):getPosition())
  finishRadius = getSceneWaypointRadius(scenetree.findObject(event.waypoints[#event.waypoints]))
  if #racePath == 0 then
    createRacePath(startPos)
  end
  setupPlayer()
	setupRacers()
  hornFlag = false
  stats.racersCount = #racers
  aiLua = aiLua.."."
  for i, v in ipairs(racers) do
    v.obj:queueLuaCommand(aiLua.."setMode('stop')")
    v.obj:queueLuaCommand(aiLua.."setAggression(1.4)")
    v.obj:queueLuaCommand("controller.setFreeze(1)")
  end
end

local function onCountdownStarted()
  hornFlag = true
  for i, v in ipairs(racers) do
    if not v.script then
      v.obj:queueLuaCommand(aiLua.."driveUsingPath({wpTargetList = "..serialize(event.waypointsAi)..", noOfLaps = "..event.lapCountAi..", aggression = "..tostring(math.min(1.4, v.aggro * 1.2))..", avoidCars = "..tostring(event.avoidCars).."})")
    end
  end
end

local function onRaceStart()
  player.state = "race"
  for i, v in ipairs(racers) do
    if v.script then
      local json = jsonReadFile(v.script)
      if json then
        v.obj:queueLuaCommand(aiLua.."startFollowing("..serialize(json.recording)..")")
      end
    end
    v.obj:queueLuaCommand("controller.setFreeze(0)")
    v.state = "startBoost"
    -- v.obj:queueLuaCommand("ai.setVehicleDebugMode({debugMode = 'speeds'})")
  end
  scoreStats.initialiseArbitraryStat("score", "Score", player.obj, player.obj.name, 5, 5)
end

local function onRaceTick(tick)
	player.vel = map.objects[player.id].vel:length()
	player.vec = vec3(player.obj:getDirectionVector())
  player.vecUp = vec3(player.obj:getDirectionVectorUp())
	player.damage = map.objects[player.id].damage or 0
  stats.vecUpDot = player.vecUp:dot(up)
  if player.vel > stats.topSpeed and player.vel <= 100 then
    stats.topSpeed = player.vel
  end
  if player.damage > stats.topDamage then
    stats.topDamage = player.damage
  end
  if stats.topSpeed <= event.crawlSpeed then
    stats.headStart = mainTimer
  end
  if stats.vecUpDot < stats.maxRoll then
    stats.maxRoll = stats.vecUpDot
  end
  if player.timers.flip >= event.stuckTime then
    scenario.finish({failed = event.messages.flip})
  end
  if player.timers.stuck >= event.stuckTime then
    scenario.finish({failed = event.messages.stuck})
  end
  if player.vel <= event.crawlSpeed then
    if stats.vecUpDot <= 0.1 then
      player.timers.flip = player.timers.flip + tick
    else
      player.timers.flip = 0
    end
    if player.timers.flip == 0 and player.damage >= event.damageCutoff then
      player.timers.stuck = player.timers.stuck + tick
    end
  else
    player.timers.stuck = 0
  end
	for i, v in ipairs(racers) do
		v.vel = map.objects[v.id].vel:length()
		v.vec = vec3(v.obj:getDirectionVector())
    v.vecUp = vec3(v.obj:getDirectionVectorUp())
		v.damage = map.objects[v.id].damage
    v.dist = v.pos:distance(player.pos)
    if v.vecUp:dot(up) <= 0 and v.dist <= 150 then
      if player.vec:dot((v.pos - player.pos):normalized()) > 0 then
        if not tableContains(stats.rolledRacers, v.id) then
          table.insert(stats.rolledRacers, v.id)
        end
      end
    end
    if (v.state ~= "start" and v.state ~= "finish") then
      if v.vel <= event.crawlSpeed then
        if not tableContains(stats.stuckRacers, v.id) then
          table.insert(stats.stuckRacers, v.id)
        end
      else
        if tableContains(stats.stuckRacers, v.id) then
          table.remove(stats.stuckRacers, arrayFindValueIndex(stats.stuckRacers, v.id))
        end
      end
    end
    if event.aggroRubberband > 0 then
      v.aggro = v.aggroBase + clamp((player.path.percent - v.path.percent) / (25 / event.aggroRubberband), -0.5, 0.05)
    end
    pathProgress(v)
	end
  pathProgress(player)
  mainTimer = mainTimer + tick
end

local function onRaceResult()
  if player.state == "finish" then
    stats.finishTime = scenario.getScenario().timer
    stats.finishTimeStr = timeToString(stats.finishTime)
    local finishText = ""
    local errorText = "Somehow, the race results glitched and are invalid. Sorry about this. Try another race!"
    local reward = 0
    local medalScore = 0
    local saveScore = 0
    for i, v in ipairs(finishers) do
      if v == "Player" then
        stats.finishPlace = i
        medalScore = math.max(1, 6 - i)
        saveScore = math.max(1, 11 - i)
      end
    end
    scoreStats.setStatProgress(player.obj:getID(), "score", player.obj.name, {value = medalScore})
    if finishers[1] then
      if finishers[1] == "Player" then
        finishText = event.messages.win
      else
        finishText = finishers[1].." "..event.messages.finish
        finishText = finishText.." "..event.messages.place.." "..ranks[stats.finishPlace].."."
      end
      if event.rewards[stats.finishPlace] then
        reward = event.rewards[stats.finishPlace]
      end
      local json = jsonReadFile(event.saveFile)
      if not json or json.series ~= event.series then
        json = setupSaveFile()
      end
      json.money = json.money + reward
      json.raceScores[event.index] = math.max(saveScore, json.raceScores[event.index])
      if saveScore == 10 then
        json.wins = json.wins + 1
      else
        stats.headStart = 0
      end
      stats.money = json.money
      stats.wins = json.wins
      if json.allowAchievements then
        stats.totalScore = 0
        stats.minScore = 10
        stats.stuckRacersCount = #stats.stuckRacers
        stats.rolledRacersCount = #stats.rolledRacers
        for i, v in ipairs(json.raceScores) do
          stats.totalScore = stats.totalScore + v
          if v < stats.minScore then
            stats.minScore = v
          end
        end
        testAchievements(json)
      end
      finishText = finishText.." Your final time is "..stats.finishTimeStr.."."
      finishText = finishText.." Your reward is $"..tostring(reward).."."
      finishText = finishText.." Your total cash is $"..tostring(json.money).."."
      if json.allowAchievements then
        finishText = finishText.." Your achievement progress is "..tostring(stats.achievementsDone).."/"..tostring(stats.achievementsTotal).." ."
        if stats.achievementsDone >= stats.achievementsTotal then
          finishText = finishText.." Perfect!"
        end
      end
      if json.money >= 1000000 then
        finishText = finishText.." It looks like you cheated to get that much money. Yoink, it's all gone now!"
        json.money = 0
      elseif json.money < 0 then
        finishText = finishText.." How the heck are you in debt? Here, you're back at $0 total now."
        json.money = 0
      end
      jsonWriteFile(event.saveFile, json)
    end
    if stats.finishPlace == 0 then
      scenario.finish({failed = errorText})
    elseif stats.finishPlace <= 3 then
      scenario.finish({msg = finishText})
    else
      scenario.finish({failed = finishText})
    end
  end
end

local function onRaceWaypointReached(data)
  if data.waypointName == event.lapPoint and data.vehicleId == player.id then
    player.lapsDone = player.lapsDone + 1
    if player.lapsDone >= event.lapCount then
      table.insert(finishers, "Player")
      player.state = "finish"
    end
  end
end

local function onPreRender(dt)
  if spawnDelay and spawnDelay == 3 then
    spawnDelay = nil
    spawnGroup()
  end
  if spawnDelay and spawnDelay >= 1 then
    spawnDelay = spawnDelay + 1
  end
  if player.obj then
    player.pos = vec3(player.obj:getPosition())
    player.obj:queueLuaCommand("obj:queueGameEngineLua('scenario_raceExtras.setHornValue('..tostring(electrics.values.horn)..')')")
    if hornValue == 1 and not hornFlag then
      local str = "No save data found."
      local json = jsonReadFile(event.saveFile)
      if json then
        if not json.raceScores[event.index] or json.raceScores[event.index] == 0 then
          str = "You haven't completed this race yet."
        else
          str = "Your best result in this race is "..ranks[11 - json.raceScores[event.index]].."."
        end
      end
      local data = {}
      data.type = "info"
      data.desc = str
      table.insert(textBoxQueue, data)
      hornFlag = true
    end
  end
  if #racers > 0 then
    for i, v in ipairs(racers) do
      if event.aiLua ~= "ai" then
        v.obj:queueLuaCommand(aiLua.."updateGFX("..tostring(dt)..")")
      end
      v.pos = vec3(v.obj:getPosition())
      if v.state == "startBoost" then
        v.distTarget = v.pos:distance(startPos)
        if v.distTarget >= event.startBoostDist then
          v.obj:queueLuaCommand(aiLua.."setAggression("..tostring(v.aggro)..")")
          v.state = "race"
        end
      else
        v.distTarget = v.pos:distance(finishPos)
        if v.state ~= "finish" and v.distTarget <= finishRadius then
          v.obj:queueLuaCommand(aiLua.."setAggression(0.25)")
          local str = v.driver or "Racer #"..tostring(v.id)
          table.insert(finishers, str)
          v.state = "finish"
        end
      end
    end
  end
  if #textBoxQueue > 0 then
    windowOpen[0] = textBoxTimer >= 0.25 and true or false
    showTextBox(textBoxQueue[1])
    if textBoxTimer >= event.textBoxTime then
      table.remove(textBoxQueue, 1)
      textBoxTimer = 0
    end
    textBoxTimer = textBoxTimer + dt
  end
end

local function onSerialize()
  local data = scenario.getScenario()
  return data
end

local function onDeserialized(data)
  event.waypoints = data.lapConfig
  event.lapCount = data.lapCount or 1
  event.lapPoint = data.lapConfig[#data.lapConfig]
  if data.raceExtras then
    for k, v in pairs(data.raceExtras) do
      if type(v) ~= "nil" then
        event[k] = v
      end
    end
  end
end

M.initScenario = initScenario
M.setHornValue = setHornValue
M.onScenarioUIReady = onScenarioUIReady
M.onCountdownStarted = onCountdownStarted
M.onRaceStart = onRaceStart
M.onRaceTick = onRaceTick
M.onRaceResult = onRaceResult
M.onRaceWaypointReached = onRaceWaypointReached
M.onPreRender = onPreRender
M.onSerialize = onSerialize
M.onDeserialized = onDeserialized

return M